
import ../types
import ./texture_decode
from dds_ktx import KtxInfo, KtxPart, get_ASTC_internal_format
import arr_ref
import stb_image_resize
import std/monotimes
import std/bitops

# TODO: don't import it here
from ../platform/gl import nil

when defined(myouUseBC7Encoder):
    import bc7enc
when defined(myouUseAstcEncoder):
    import astc
when defined(android) or defined(ios) or defined(emscripten):
    template has_bptc_support: bool = gl.GLAD_GL_EXT_texture_compression_bptc
else:
    template has_bptc_support: bool = gl.GLAD_GL_ARB_texture_compression_bptc
template has_astc_support: bool = gl.GLAD_GL_OES_texture_compression_astc or
                                  gl.GLAD_GL_KHR_texture_compression_astc_ldr

const myouEngineNumTextureThreads {.intdefine.} = 4
const myouEngineCompressTextures {.booldefine.} = true
const myouBC7VerboseMode {.booldefine.} = true

template u32(x: untyped): uint32 = cast[uint32](x)

type CallbackUncompressed = proc(tex: Texture, data: SliceMem[byte]) {.gcsafe.}
type CallbackCompressed = proc(tex: Texture, info: KtxInfo, data: seq[KtxPart], refdata: seq[ArrRef[byte]]) {.gcsafe.}
type CompressMipmapResult = tuple[data: ArrRef[byte], row_len: int]
type CompressMipmap = proc(width, height: int32, p: pointer, len: int32): CompressMipmapResult {.closure.}

template block_file_size(width, height, depth, block_x, block_y, block_z, block_byte_size: untyped): int = 
    let blocks_x = (width.int + block_x.int - 1) div block_x.int
    let blocks_y = (height.int + block_y.int - 1) div block_y.int
    let blocks_z = (depth.int + block_z.int - 1) div block_z.int
    blocks_x * blocks_y * blocks_z * block_byte_size.int

proc make_mipmaps(tex: Texture, pixels: SliceMem[byte], compress: CompressMipmap): (seq[KtxPart], seq[ArrRef[byte]]) =
    var width = tex.width.int32
    var height = tex.height.int32
    let depth = tex.format_depth
    assert depth == 1, "Compressed 2D arrays and 3D textures not supported yet"
    assert tex.tex_type != TexCube, "Compressed cube textures not supported yet"
    let stbir_data_type = case tex.format.component_type:
        of UByte:
            if tex.is_sRGB: STBIR_TYPE_UINT8_SRGB
            else: STBIR_TYPE_UINT8
        of HalfFloat: STBIR_TYPE_HALF_FLOAT
        of Float: STBIR_TYPE_FLOAT
        of UShort: STBIR_TYPE_UINT16
        else: raise Defect.newException "unreachable"
    # 4 because we're forcing 4 channels for the encoder
    let pixel_stride = 4 * (tex.format.stride div tex.format.channel_count)
    let time = getmonotime().ticks.float/1000000000
    var parts: seq[KtxPart]
    var data_refs: seq[ArrRef[byte]]
    var mip_level = 0'i32
    var w, prev_w = width
    var h, prev_h = height
    # round down to POT (TODO: make configurable)
    w = 1'i32 shl fastLog2(prev_w)
    h = 1'i32 shl fastLog2(prev_h)
    let resize_level_0 = w != prev_w or h != prev_h
    var current_level, last_level: ArrRef[byte]
    var last_src: pointer = pixels.data
    var last_len: int
    while w >= 4 or h >= 4 or parts.len == 0:
        (last_src, last_len) = if mip_level != 0 or resize_level_0:
            current_level = newArrRef[byte](w*h*pixel_stride)
            stbir_resize(
                last_src, prev_w, prev_h, 0,
                current_level.toPointer, w, h, 0,
                StbirRgba, stbir_data_type,
                STBIR_EDGE_CLAMP, # TODO: handle other cases
                STBIR_FILTER_BOX,
            )
            last_level = current_level
            (current_level.toPointer, current_level.byte_len.int)
        else:
            (pixels.toPointer, pixels.byte_len)
        
        let (data, row_len) = compress(w, h, last_src, last_len.int32)

        parts.add KtxPart(
            width: w, height: h, data: data.toPointer,
            len: data.byte_len, mip_level: mip_level,
            row_len: row_len,
        )
        data_refs.add data
        prev_w = w; prev_h = h
        w = max(1, w shr 1)
        h = max(1, h shr 1)
        mip_level.inc
    let time2 = getmonotime().ticks.float/1000000000
    echo "time: ", time2-time, " ", tex.name
    return (parts, data_refs)

when defined(myouUseBC7Encoder):
    proc bcn_compress*(tex: Texture, pixels: SliceMem[byte], callback: CallbackCompressed, bc_format: int8) = 
        when not defined(android):
            let (bpp, internal_format) = get_bc_bpp_internal_format(bc_format, tex.is_sRGB)
            let block_size_bytes = (4*4*bpp.int) div 8
            proc compress(w, h: int32, p: pointer, len: int32): CompressMipmapResult {.closure.} =
                var input = EncodeBcInput(
                    data: p, len: len,
                    width: w, height: h,
                    format: bc_format,
                )
                let out_len = block_file_size(w,h,1, 4,4,1, block_size_bytes)
                let row_len = block_file_size(w,1,1, 4,4,1, block_size_bytes)
                var data = newArrRef[byte](out_len)
                var output = EncodeBcOutput(
                    data: data.toPointer,
                    len: out_len.int32,
                )
                let err = encode_bc(input, output, myouBC7VerboseMode)
                assert err == NO_ERROR, "bc7enc error: " & $err
                return (data, row_len)

            let (parts, data_refs) = make_mipmaps(tex, pixels, compress)

            let info = KtxInfo(
                width: parts[0].width, height: parts[0].height, depth: tex.format_depth.int32,
                num_layers: 1, num_mipmaps: parts.len.int32, has_alpha: bc_format in [2,3,7],
                is_sRGB: tex.is_sRGB, is_bc: true,
                internal_format: internal_format,
            )
            callback(tex, info, parts, data_refs)

when defined(myouUseAstcEncoder):
    proc atsc_compress*(tex: Texture, pixels: SliceMem[byte], callback: CallbackCompressed, blk_size = (6,6), quality = 10.0) =
        let profile = case tex.format.component_type:
            of UByte:
                if tex.is_sRGB: ASTCENC_PRF_LDR_SRGB
                else: ASTCENC_PRF_LDR
            of HalfFloat: ASTCENC_PRF_HDR_RGB_LDR_A
            of Float: ASTCENC_PRF_HDR_RGB_LDR_A
            of UShort: ASTCENC_PRF_HDR_RGB_LDR_A
            else: raise Defect.newException "unreachable"
        let data_type = case tex.format:
            of SRGB_u8, SRGB_Alpha_u8, R_u8, RG_u8, RGB_u8, RGBA_u8: ASTCENC_TYPE_U8
            of R_f16, RG_f16, RGB_f16, RGBA_f16: ASTCENC_TYPE_F16
            of R_f32, RG_f32, RGB_f32, RGBA_f32: ASTCENC_TYPE_F32
            else: raise ValueError.newException "Unsupported data type of " & $tex.format
        let swizzle = case tex.format.channel_count:
            # TODO: more optimal usage for 1/2 channels, normals, etc.
            of 1: AstcencSwizzle(r: SWZ_R, g: SWZ_R, b: SWZ_R, a: SWZ_1)
            of 2: AstcencSwizzle(r: SWZ_R, g: SWZ_G, b: SWZ_0, a: SWZ_1)
            of 3: AstcencSwizzle(r: SWZ_R, g: SWZ_G, b: SWZ_B, a: SWZ_1)
            else: AstcencSwizzle(r: SWZ_R, g: SWZ_G, b: SWZ_B, a: SWZ_A)
        let flags = 0'u32
        var config: AstcencConfig
        # TODO: 3D texture block size
        astc_assert astcenc_config_init(profile,
            blk_size[0].uint32, blk_size[1].uint32, 1'u32,
            quality, flags, config.addr)
        # config.progress_callback = proc(p: float32) {.cdecl.} =
        #     echo "progress ", p
        var ctx: ptr AstcencContext
        astc_assert astcenc_context_alloc(config.addr, 1, ctx)
        let time = getmonotime().ticks.float/1000000000
        proc compress(w, h: int32, p: pointer, len: int32): CompressMipmapResult {.closure.} =
            let buffer_size = block_file_size(w, h, 1,
                config.block_x, config.block_y, config.block_z, 16)
            let row_len = block_file_size(w, 1, 1,
                config.block_x, config.block_y, config.block_z, 16)
            let data = newArrRef[byte](buffer_size)
            let pointers = [p]
            let img = AstcencImage(
                dim_x: w.u32, dim_y: h.u32, dim_z: 1.u32,
                data_type: data_type,
                data: pointers[0].addr,
            )
            const thread_index = 0 # TODO
            let err = astcenc_compress_image(ctx, img.addr, swizzle.addr,
                data[0].addr, buffer_size.csize_t, thread_index)
            assert err == ASTCENC_SUCCESS, "ASTC encoding error: " & $err
            return (data, row_len)

        let (parts, data_refs) = make_mipmaps(tex, pixels, compress)
        
        let info = KtxInfo(
            width: parts[0].width, height: parts[0].height, depth: tex.format_depth.int32,
            num_layers: 1, num_mipmaps: parts.len.int32, has_alpha: tex.format.channel_count == 4,
            is_sRGB: tex.is_sRGB, is_astc: true,
            internal_format: get_ASTC_internal_format(blk_size, tex.is_sRGB)
        )
        callback(tex, info, parts, data_refs)
        astc_assert astcenc_compress_reset(ctx)
        astcenc_context_free(ctx)

proc loadOptimized*(tex: Texture, slices: seq[SliceMem[byte]],
        callback_uncompressed: CallbackUncompressed = nil,
        callback_compressed: CallbackCompressed = nil,
        flip = true, min_channels = 0) {.gcsafe.} =
    
    var min_channels = min_channels
    var will_compress = myouEngineCompressTextures
    var will_load_uncompressed_first = false
    var will_encode_all = false

    will_compress = will_compress and
        callback_compressed != nil and
        tex.format.component_type != UShort
    
    if has_bptc_support:
        # TODO: BC6H or RGBM
        will_compress = will_compress and
            tex.format.component_type == UByte
    elif has_astc_support:
        # TODO: detect HDR support
        discard
    elif will_encode_all:
        will_load_uncompressed_first = true
    else:
        will_compress = false
    
    if will_compress:
        min_channels = 4
    
    tex.loadFileFromSlices(slices, proc(tex: Texture, pixels: SliceMem[byte]) {.gcsafe.} =
        if callback_uncompressed != nil and 
                (will_load_uncompressed_first or not will_compress):
            callback_uncompressed(tex, pixels)
        if will_compress:
            when defined(myouUseBC7Encoder):
                let bc_format = if tex.format.channel_count == 1:
                    4.int8 # BC4 is 0.5 bytes per pixel grayscale
                else:
                    7.int8 # BC7 is just the best at anything else
                if has_bptc_support:
                    bcn_compress(tex, pixels, callback_compressed, bc_format)
                    if not will_encode_all: return
                elif will_encode_all:
                    # TODO: callback to store result
                    discard
                    # bcn_compress(tex, pixels, ..., bc_format)
            when defined(myouUseAstcEncoder):
                if has_astc_support or will_encode_all:
                    atsc_compress(tex, pixels, callback_compressed)
                    if not will_encode_all: return
                elif will_encode_all:
                    # TODO: callback to store result
                    discard
                    # atsc_compress(tex, pixels, ...)

    , flip = flip, min_channels = min_channels)


# main -> thread  channels
type DecodeChanMsg = tuple[
    tex: Texture, slices: seq[SliceMem[byte]],
    callback_uncompressed: CallbackUncompressed,
    callback_compressed: CallbackCompressed,
    flip: bool, min_channels: int,
]
var decode_chan: Channel[DecodeChanMsg]
# main <- thread  channels
type DecodeReturnChanMsg = tuple[
    callback: CallbackUncompressed,
    tex: Texture, data: SliceMem[byte],
]
var decode_return_chan: Channel[DecodeReturnChanMsg]
type CompressedReturnChanMsg = tuple[
    callback: CallbackCompressed,
    tex: Texture, info: KtxInfo, data: seq[KtxPart], refdata: seq[ArrRef[byte]],
]
var compressed_return_chan: Channel[CompressedReturnChanMsg]

proc loadOptimizedThreaded*(tex: Texture, slices: seq[SliceMem[byte]],
        callback_uncompressed: CallbackUncompressed = nil,
        callback_compressed: CallbackCompressed = nil,
        flip = true, min_channels = 0) =
    
    when false:
        loadOptimized(tex, slices, callback_uncompressed, callback_compressed, flip, min_channels)
    else:
        decode_chan.send((tex: tex, slices: slices,
                    callback_uncompressed: callback_uncompressed,
                    callback_compressed: callback_compressed,
                    flip: flip, min_channels: min_channels))

var workers = newSeq[Thread[void]](myouEngineNumTextureThreads)
proc workerThreadProc() {.thread.} =
    # TODO: handle errors
    while true:
        let to_decode = decode_chan.recv()
        if to_decode.tex == nil:
            break
        proc cb(tex: Texture, data: SliceMem[byte]) =
            decode_return_chan.send((callback: to_decode.callback_uncompressed, tex: tex, data: data))
        proc cbc(tex: Texture, info: KtxInfo, data: seq[KtxPart], refdata: seq[ArrRef[byte]]) =
            compressed_return_chan.send((callback: to_decode.callback_compressed, tex: tex, 
            info: info, data: data, refdata: refdata))
        let cb_out = if to_decode.callback_uncompressed != nil: cb else: nil
        let cbc_out = if to_decode.callback_compressed != nil: cbc else: nil
        loadOptimized(to_decode.tex, to_decode.slices, cb_out, cbc_out,
                    to_decode.flip, to_decode.min_channels)

decode_chan.open()
decode_return_chan.open()
compressed_return_chan.open()
for worker in workers.mitems:
    worker.createThread(workerThreadProc)

proc updateTextureWorkerThreads*() =
    # TODO: handle errors
    while true:
        let tried = decode_return_chan.tryRecv()
        if not tried.dataAvailable:
            break
        let (cb, tex, data) = tried.msg
        cb(tex, data)
    while true:
        let tried = compressed_return_chan.tryRecv()
        if not tried.dataAvailable:
            break
        let (cb, tex, info, data, refdata) = tried.msg
        cb(tex, info, data, refdata)

proc terminateTextureWorkerThreads*() =
    for worker in workers:
        decode_chan.send(DecodeChanMsg.default)
    for worker in workers:
        worker.joinThread()
